@@ -1,26 +1,43 @@
 import logging
 
-from ckeditor.fields import RichTextField
 from django.apps import apps
 from django.contrib.auth import get_user_model
+from django.contrib.auth.views import redirect_to_login
+from django.contrib.contenttypes.fields import GenericForeignKey
+from django.contrib.contenttypes.models import ContentType
 from django.contrib.sites.models import Site
 from django.core.exceptions import ValidationError
 from django.db import models
-from django.urls import NoReverseMatch, get_script_prefix, reverse
+from django.http.response import Http404, HttpResponse
+from django.template import loader
+from django.urls import reverse
 from django.utils import timezone
-from django.utils.encoding import iri_to_uri
 from django.utils.translation import gettext_lazy as _
+from django.views.generic import ListView
+
+from ckeditor.fields import RichTextField
 from filer.fields.image import FilerImageField
 from mptt.models import MPTTModel, TreeForeignKey
 from simpel_themes.models import ModelTemplate
 from taggit.managers import TaggableManager
 from taggit.models import TagBase, TaggedItemBase
 
-from .utils import unique_slugify
+from simpel_pages.managers import CategoryManager, PageManager, TagManager
+from simpel_pages.utils import unique_slugify
+
+from .routable import RouteResult
+from .settings import pages_settings
 
 logger = logging.getLogger(__name__)
 
 
+class IndexView(ListView):
+    page_template = None
+
+    def render_to_response(self, context, **response_kwargs):
+        return HttpResponse(self.page_template.render(context, self.request))
+
+
 class Category(MPTTModel):
 
     parent = TreeForeignKey(
@@ -35,11 +52,22 @@ class Category(MPTTModel):
             " and Big Band. Totally optional."
         ),
     )
+    thumbnail = FilerImageField(
+        null=True,
+        blank=True,
+        on_delete=models.SET_NULL,
+        related_name="pagecategories",
+    )
     name = models.CharField(
         max_length=80,
         unique=True,
         verbose_name=_("Category Name"),
     )
+    description = models.TextField(
+        null=True,
+        blank=True,
+        verbose_name=_("Description"),
+    )
     slug = models.SlugField(
         unique=True,
         null=True,
@@ -47,7 +75,10 @@ class Category(MPTTModel):
         editable=False,
         max_length=80,
     )
+    created_at = models.DateTimeField(default=timezone.now)
+    last_modified_at = models.DateTimeField(default=timezone.now)
 
+    objects = CategoryManager()
     icon = "tag-outline"
 
     class Meta:
@@ -66,6 +97,9 @@ class Category(MPTTModel):
     def opts(self):
         return self.__class__._meta
 
+    def get_absolute_url(self):
+        return reverse("serve_category", kwargs={"slug": self.slug})
+
     def clean(self):
         if self.parent:
             parent = self.parent
@@ -82,8 +116,17 @@ class Category(MPTTModel):
 
 class Tag(TagBase):
 
+    description = models.TextField(
+        null=True,
+        blank=True,
+        verbose_name=_("Description"),
+    )
+    created_at = models.DateTimeField(default=timezone.now)
+    last_modified_at = models.DateTimeField(default=timezone.now)
     icon = "tag-outline"
 
+    objects = TagManager()
+
     class Meta:
         verbose_name = _("Tag")
         verbose_name_plural = _("Tags")
@@ -92,6 +135,9 @@ class Tag(TagBase):
     def opts(self):
         return self._meta
 
+    def get_absolute_url(self):
+        return reverse("serve_tag", kwargs={"slug": self.slug})
+
 
 class Page(MPTTModel):
     created_at = models.DateTimeField(
@@ -102,17 +148,12 @@ class Page(MPTTModel):
         default=timezone.now,
         verbose_name=_("Last updated at"),
     )
-    sites = models.ManyToManyField(
-        Site,
-        verbose_name=_("sites"),
-        related_name="simpelposts",
-    )
     owner = models.ForeignKey(
         get_user_model(),
         null=True,
         blank=True,
         on_delete=models.SET_NULL,
-        related_name="%(class)ss",
+        related_name="pages",
         verbose_name=_("Page Owner"),
     )
     parent = TreeForeignKey(
@@ -120,28 +161,48 @@ class Page(MPTTModel):
         blank=True,
         null=True,
         on_delete=models.SET_NULL,
-        related_name="children",
+        related_name="childs",
         help_text=_(
             "Pages, unlike tags, can have a hierarchy. You might have a "
             "Index page, and under that have children pages for post"
             " and story. Totally optional."
         ),
     )
-    url = models.CharField(
-        _("URL"),
-        max_length=100,
-        db_index=True,
-    )
     title = models.CharField(
         _("title"),
         max_length=200,
     )
+    category = TreeForeignKey(
+        Category,
+        related_name="pages",
+        on_delete=models.SET_NULL,
+        null=True,
+        blank=True,
+        verbose_name=_("Category"),
+    )
+    tags = TaggableManager(
+        through="TaggedPage",
+        blank=True,
+        related_name="pages",
+        verbose_name=_("Tags"),
+    )
+    thumbnail = FilerImageField(
+        null=True,
+        blank=True,
+        on_delete=models.SET_NULL,
+        related_name="simpelpages",
+    )
+    content = RichTextField(
+        null=True,
+        blank=True,
+        verbose_name=_("Content"),
+    )
+    data = models.JSONField(null=True, blank=True)
     slug = models.SlugField(
         unique=True,
         null=True,
         blank=True,
         db_index=True,
-        editable=False,
         max_length=255,
     )
     seo_title = models.CharField(
@@ -160,10 +221,28 @@ class Page(MPTTModel):
         on_delete=models.SET_NULL,
         null=True,
         blank=True,
-        related_name="pages",
+        related_name="%(class)s_templates",
         verbose_name=_("Template"),
     )
-
+    url_path = models.CharField(
+        max_length=255,
+        null=True,
+        blank=True,
+    )
+    index = models.BooleanField(
+        _("index"),
+        default=False,
+        help_text=_("Index page, contain child pages and not included in widgets."),
+    )
+    per_page = models.IntegerField(
+        default=0,
+        verbose_name=_("per page"),
+        help_text=_("Number item per page, used only when index is True."),
+    )
+    live = models.BooleanField(
+        _("live"),
+        default=True,
+    )
     allow_comments = models.BooleanField(
         _("allow comments"),
         default=False,
@@ -173,14 +252,12 @@ class Page(MPTTModel):
         help_text=_("If this is checked, only logged-in users will be able to view the page."),
         default=False,
     )
-    data = models.JSONField(null=True, blank=True)
     real_model = models.CharField(
         max_length=120,
         editable=False,
         null=True,
         blank=True,
     )
-
     readers = models.ManyToManyField(
         get_user_model(),
         blank=True,
@@ -193,43 +270,111 @@ class Page(MPTTModel):
         related_name="bookmarks",
         verbose_name=_("Users who bookmark this page."),
     )
-    category = models.ForeignKey(
-        "simpel_pages.Category",
-        related_name="pages",
-        on_delete=models.SET_NULL,
-        null=True,
-        blank=True,
-        verbose_name=_("Category"),
-    )
-    tags = TaggableManager(
-        through="TaggedPage",
-        blank=True,
-        related_name="pages",
-        verbose_name=_("Tags"),
-    )
+    visitor_count = models.IntegerField(default=0)
+
+    objects = PageManager()
+
+    def get_descendants(self, include_self=False):
+        return super().get_descendants(include_self)
 
     class Meta:
-        verbose_name = _("page")
-        verbose_name_plural = _("pages")
-        ordering = ["url"]
+        ordering = ["-created_at"]
+
+    def __str__(self):
+        return "%s" % self.title
 
     @property
     def opts(self):
         return self.get_real_model_class()._meta
 
-    def __str__(self):
-        return "%s -- %s" % (self.url, self.title)
+    @property
+    def specific(self):
+        return self.get_real_instance()
 
-    def get_absolute_url(self):
-        from .views import simpelpage
+    @property
+    def page_type(self):
+        return self.opts.verbose_name
+
+    def get_related_by_category(self, number=5):
+        return self.__class__.objects.get_related_by_category(self)[:number]
+
+    def get_related_by_tags(self, number=5):
+        return self.__class__.objects.get_related_by_tags(self)[:number]
+
+    def has_permission(self):
+        return True
+
+    def route(self, request, path_components):
+        if path_components:
+            # request is for a child of this page
+            child_slug = path_components[0]
+            remaining_components = path_components[1:]
 
-        for url in (self.url.lstrip("/"), self.url):
             try:
-                return reverse(simpelpage, kwargs={"url": url})
-            except NoReverseMatch:
-                pass
-        # Handle script prefix manually because we bypass reverse()
-        return iri_to_uri(get_script_prefix().rstrip("/") + self.url)
+                subpage = self.get_children().get(slug=child_slug)
+            except self.__class__.DoesNotExist:
+                raise Http404
+
+            return subpage.specific.route(request, remaining_components)
+
+        else:
+            # request is for this very page
+            if self.live:
+                return RouteResult(self)
+            else:
+                raise Http404
+
+    def get_context_data(self, request, **kwargs):
+        context = {
+            "object": self,
+            "page_title": self.seo_title or self.title,
+            "page_description": self.seo_description or self.title,
+        }
+        context.update(kwargs)
+        return context
+
+    def get_template(self, request, *args, **kwargs):
+        if not self.template:
+            template = loader.get_template(pages_settings.DEFAULT_TEMPLATE)
+        else:
+            template = self.template.specific
+        return template
+
+    def serve_index(self, request, *args, **kwargs):
+        context = self.get_context_data(request, **kwargs)
+        constructor = {
+            "paginate_by": self.per_page or pages_settings.ITEM_PER_PAGE,
+            "page_template": self.get_template(request),
+            "extra_context": context,
+            "queryset": self.childs.live(),
+            "response_class": HttpResponse,
+        }
+        return IndexView.as_view(**constructor)(request, *args, **kwargs)
+
+    def serve_page(self, request, *args, **kwargs):
+        template = self.get_template(request, *args, **kwargs)
+        context = self.get_context_data(request, **kwargs)
+        return HttpResponse(template.render(context, request))
+
+    def serve(self, request, *args, **kwargs):
+        if self.registration_required and not request.user.is_authenticated:
+            return redirect_to_login(request.path)
+        if self.index:
+            return self.serve_index(request, *args, **kwargs)
+        return self.serve_page(request, *args, **kwargs)
+
+    def set_url_path(self):
+        if self.parent is not None:
+            self.url_path = self.parent.url_path + self.slug + "/"
+        else:
+            self.url_path = ""
+        return self.url_path
+
+    def page_url(self, site=None, request=None):
+        return reverse("serve_pages", args=(self.url_path,))
+
+    def get_absolute_url(self, site=None, request=None):
+        return self.page_url(site, request)
 
     def get_model_name(self):
         return "%s.%s" % (self.opts.app_label, self.opts.model_name)
@@ -272,30 +417,35 @@ class Page(MPTTModel):
             unique_slugify(self, self.title)
         if not self.real_model:
             self.real_model = self.get_model_name()
+        self.set_url_path()
         return super().save(*args, **kwargs)
 
 
-class Post(Page):
-    thumbnail = FilerImageField(
-        null=True,
-        blank=True,
+class RootPage(models.Model):
+    site = models.OneToOneField(
+        Site,
         on_delete=models.SET_NULL,
-        related_name="simpelpages",
-    )
-    content = RichTextField(
-        _("content"),
         null=True,
         blank=True,
+        related_name="rootpage",
+        verbose_name=_("Site"),
     )
-    summary = RichTextField(
-        _("summary"),
+    page = models.OneToOneField(
+        Page,
+        on_delete=models.SET_NULL,
         null=True,
+        related_name="site",
         blank=True,
+        verbose_name=_("Page"),
     )
 
-    class Meta(Page.Meta):
-        verbose_name = _("post")
-        verbose_name_plural = _("posts")
+    class Meta:
+        verbose_name = _("Root")
+        verbose_name_plural = _("Roots")
+        swappable = "SIMPEL_PAGE_SITE_MODEL"
+
+    def __str__(self):
+        return "%s root Page" % self.site
 
 
 class TaggedPage(TaggedItemBase):
@@ -303,7 +453,7 @@ class TaggedPage(TaggedItemBase):
     content_object = models.ForeignKey(
         Page,
         on_delete=models.CASCADE,
-        related_name="tagged_pages",
+        related_name="tagged_posts",
         db_index=True,
     )
     tag = models.ForeignKey(
@@ -321,34 +471,17 @@ class TaggedPage(TaggedItemBase):
         return str(self.tag)
 
 
-class PageGallery(models.Model):
-    page = models.ForeignKey(
-        Page,
-        on_delete=models.CASCADE,
-        related_query_name="galleries",
-        verbose_name=_("page"),
-    )
-    caption = models.CharField(
-        max_length=200,
-        verbose_name=_("Caption"),
-    )
-    thumb_height = models.IntegerField(
-        default=100,
-        verbose_name=_("Thumbnail Height"),
-    )
-    thumb_width = models.IntegerField(
-        default=100,
-        verbose_name=_("Thumbnail Width"),
-    )
-    image = FilerImageField(
+class Visitor(models.Model):
+    ip = models.CharField(max_length=16, null=True)
+    url = models.CharField(max_length=255)
+    path = models.CharField(max_length=255)
+    referrer = models.CharField(max_length=255, null=True)
+    created_at = models.DateTimeField(default=timezone.now)
+    content_type = models.ForeignKey(
+        ContentType,
+        on_delete=models.SET_NULL,
         null=True,
         blank=True,
-        on_delete=models.SET_NULL,
-        related_name="simpelpages_gallery_items",
     )
-
-    class Meta:
-        verbose_name = _("Image Gallery")
-        verbose_name_plural = _("Image Galleries")
-        index_together = ("page", "image")
-        ordering = ["page"]
+    content_id = models.CharField(max_length=255, null=True)
+    content = GenericForeignKey("content_type", "content_id")
