@@ -9,4 +9,4 @@ class SimpelPageSitemap(Sitemap):
             raise ImproperlyConfigured("SimpelPageSitemap requires django.contrib.sites, which isn't installed.")
         Site = django_apps.get_model("sites.Site")
         current_site = Site.objects.get_current()
-        return current_site.simpelpages.filter(registration_required=False)
+        return current_site.simpelpages.filter(registration_required=False, live=True)
